from __future__ import annotations

from typing import Any

import torch.nn as nn  # noqa: PLR0402
from fouroversix.utils import AdaptiveBlockScalingRule, FP4Format

from .fp4_linear import FP4Linear


def quantize_model(
    model: nn.Module,
    *,
    exclude_layers: list[str] | None = None,
    fp4_format: FP4Format = FP4Format.nvfp4,
    a_scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
    w_scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
    w_scale_2d: bool = False,
    linear_cls: type[FP4Linear] | None = None,
    **kwargs: dict[str, Any],
) -> None:
    if exclude_layers is None:
        exclude_layers = ["lm_head"]

    if linear_cls is None:
        linear_cls = FP4Linear

    for name, module in model.named_modules():
        if name in exclude_layers or not isinstance(module, nn.Linear):
            continue

        fp4_linear = linear_cls(
            module.in_features,
            module.out_features,
            module.bias is not None,
            device=module.weight.device,
            dtype=module.weight.dtype,
            fp4_format=fp4_format,
            a_scale_rule=a_scale_rule,
            w_scale_rule=w_scale_rule,
            w_scale_2d=w_scale_2d,
            **kwargs,
        )

        fp4_linear.weight = module.weight
        fp4_linear.bias = module.bias
        fp4_linear.apply_ptq()

        model.set_submodule(name, fp4_linear)



from __future__ import annotations

from enum import Enum
from typing import Any

import torch
import torch.nn.functional as F

from .quantize import FP4Tensor
from .utils import AdaptiveBlockScalingRule, FP4Format, RoundStyle

SM_100 = 10
SM_110 = 11
SM_120 = 12


class MatmulBackend(str, Enum):
    """
    Backends for matrix multiplication with FP4.

    - `cutlass`: CUTLASS implementation. This requires a Blackwell GPU.
    - `pytorch`: PyTorch implementation which first dequantizes the input tensors to
        BF16 and then performs a BF16 matrix multiplication.
    """

    cutlass = "cutlass"
    pytorch = "pytorch"

    @classmethod
    def auto_select(cls) -> MatmulBackend:
        """Select the fastest backend for the given parameters."""

        for backend in [cls.cutlass, cls.pytorch]:
            if backend.is_available():
                return backend

        msg = "No available backend found"
        raise ValueError(msg)

    def is_available(self) -> bool:
        """Check if the backend is available given the CUDA device and installation."""

        if self == MatmulBackend.cutlass:
            return torch.cuda.is_available() and torch.cuda.get_device_capability()[
                0
            ] in [SM_100, SM_110, SM_120]

        return True

    def fp4_matmul(
        self,
        input: FP4Tensor,
        other: FP4Tensor,
        *,
        out_dtype: torch.dtype,
    ) -> torch.Tensor:
        """
        Perform a matrix multiplication with two FP4-quantized tensors. See frontend.py
        for more details.
        """

        if input.fp4_format != other.fp4_format:
            msg = "Both inputs must have the same FP4 format"
            raise ValueError(msg)

        if input.original_shape[1] != other.original_shape[1]:
            msg = (
                "The first input must be in row-major layout, the second input must be"
                "in column-major layout, and both inputs must have the same inner "
                "dimension"
            )
            raise ValueError(msg)

        out_shape = (input.original_shape[0], other.original_shape[0])

        if input.fp4_format == FP4Format.mxfp4:
            alpha = torch.ones(
                1,
                device=input.e2m1_values.device,
                dtype=torch.float32,
            )
        elif input.fp4_format == FP4Format.nvfp4:
            alpha = (
                (input.amax * other.amax)
                / (
                    input.scale_rule.max_allowed_e2m1_value()
                    * input.scale_rule.max_allowed_e4m3_value()
                    * other.scale_rule.max_allowed_e2m1_value()
                    * other.scale_rule.max_allowed_e4m3_value()
                )
            ).to(torch.float32)

        if self == MatmulBackend.cutlass:
            from .ops import (
                gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt,
                gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt_sm120,
                gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt,
                gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt_sm120,
                gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt,
                gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt_sm120,
            )

            gemm_fns = {
                (
                    SM_100,
                    FP4Format.mxfp4,
                    torch.bfloat16,
                ): gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt,
                (
                    SM_120,
                    FP4Format.mxfp4,
                    torch.bfloat16,
                ): gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt_sm120,
                (
                    SM_100,
                    FP4Format.nvfp4,
                    torch.bfloat16,
                ): gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt,
                (
                    SM_120,
                    FP4Format.nvfp4,
                    torch.bfloat16,
                ): gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt_sm120,
                (
                    SM_100,
                    FP4Format.nvfp4,
                    torch.float16,
                ): gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt,
                (
                    SM_120,
                    FP4Format.nvfp4,
                    torch.float16,
                ): gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt_sm120,
            }

            gemm_fn = gemm_fns.get(
                (torch.cuda.get_device_capability()[0], input.fp4_format, out_dtype),
            )

            if gemm_fn is None:
                msg = (
                    "No gemm function found for the given device capability and "
                    f"out_dtype: {torch.cuda.get_device_capability()[0]}, {out_dtype}"
                )
                raise ValueError(msg)

            out = gemm_fn(
                input.e2m1_values,
                other.e2m1_values,
                input.scale_factors,
                other.scale_factors,
                alpha,
            )

        elif self == MatmulBackend.pytorch:
            out = torch.matmul(
                input.dequantize(dtype=torch.float32),
                other.dequantize(dtype=torch.float32).T,
            ).to(out_dtype)

        else:
            msg = f"Invalid backend: {self}"
            raise ValueError(msg)

        if out_shape is not None and out.shape != out_shape:
            out = out[: out_shape[0], : out_shape[1]]

        return out


class QuantizeBackend(str, Enum):
    """
    Backends for quantizing a tensor to NVFP4 or MXFP4.

    - `cuda`: CUDA implementation. Requires a Blackwell GPU, and currently only supports
        the forward pass for PTQ (no stochastic rounding, no transposed matrices, no
        RHT, no 2D block scaling).
    - `pytorch`: PyTorch implementation.
    - `triton`: Triton implementation. Requires a Blackwell GPU.
    """

    cuda = "cuda"
    pytorch = "pytorch"
    transformer_engine = "transformer_engine"
    triton = "triton"

    @classmethod
    def auto_select(
        cls,
        x: torch.Tensor,
        **kwargs: dict[str, Any],
    ) -> QuantizeBackend:
        """Select the fastest backend for the given quantization parameters."""

        for backend in [cls.cuda, cls.triton, cls.pytorch]:
            if backend.is_supported(x, **kwargs):
                return backend

        msg = f"No backend found for the given parameters: {kwargs}"
        raise ValueError(msg)

    def is_available(self) -> bool:
        """Check if the backend can be used given the CUDA device and installation."""

        if self == QuantizeBackend.cuda:
            # TODO(jack, junxian): Re-enable CUDA backend once precision issues are
            # resolved
            return False

        if self == QuantizeBackend.triton:  # noqa: SIM102
            if not torch.cuda.is_available() or torch.cuda.get_device_capability()[
                0
            ] not in [SM_100, SM_110, SM_120]:
                return False

        if self == QuantizeBackend.transformer_engine:
            return torch.cuda.is_available()

        return True

    def is_supported(  # noqa: C901, PLR0911
        self,
        x: torch.Tensor,
        *,
        scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
        block_scale_2d: bool = False,
        had: torch.Tensor | None = None,
        fp4_format: FP4Format = FP4Format.nvfp4,
        round_style: RoundStyle = RoundStyle.nearest,
        transpose: bool = False,
    ) -> bool:
        """Check if the backend supports the given quantization parameters."""

        if x.ndim != 2:  # noqa: PLR2004
            return False

        if not self.is_available():
            return False

        if fp4_format == FP4Format.mxfp4 and scale_rule not in (
            AdaptiveBlockScalingRule.always_6,
            AdaptiveBlockScalingRule.always_4,
        ):
            msg = (
                "MXFP4 quantization only supports the `always_6` and `always_4` scale "
                "rules"
            )
            raise ValueError(msg)

        if self == QuantizeBackend.cuda:
            return (
                had is None
                and fp4_format == FP4Format.nvfp4
                and round_style == RoundStyle.nearest
                and not block_scale_2d
                and not transpose
            )

        if self == QuantizeBackend.pytorch:
            return True

        if self == QuantizeBackend.triton:
            if round_style == RoundStyle.stochastic:
                return torch.cuda.get_device_capability()[0] == SM_100

            return True

        if self == QuantizeBackend.transformer_engine:
            if (
                fp4_format != FP4Format.nvfp4
                or scale_rule != AdaptiveBlockScalingRule.always_6
            ):
                return False

            if not transpose and had is not None:
                return False

            if transpose and had is not None and block_scale_2d:  # noqa: SIM103
                return False

            return True

        msg = f"Invalid backend: {self}"
        raise ValueError(msg)

    def quantize_to_fp4(
        self,
        x: torch.Tensor,
        *,
        scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
        block_scale_2d: bool = False,
        had: torch.Tensor | None = None,
        fp4_format: FP4Format = FP4Format.nvfp4,
        round_style: RoundStyle = RoundStyle.nearest,
        transpose: bool = False,
        **kwargs: dict[str, Any],
    ) -> FP4Tensor:
        """Quantize a tensor to FP4. See frontend.py for more details."""

        original_shape = x.shape

        if transpose:
            original_shape = (original_shape[1], original_shape[0])

        if self == QuantizeBackend.cuda:
            msg = "The CUDA backend is currently disabled and will be updated soon"
            raise NotImplementedError(msg)

        elif self == QuantizeBackend.transformer_engine:
            from transformer_engine.pytorch.tensor.nvfp4_tensor import NVFP4Quantizer

            from .quantize.reference import to_blocked

            q = NVFP4Quantizer(
                with_2d_quantization=block_scale_2d,
                with_rht=had is not None,
                with_post_rht_amax=had is not None,
                stochastic_rounding=round_style == RoundStyle.stochastic,
            )

            out = q.quantize(x)

            if transpose:
                e2m1_values = out._columnwise_data  # noqa: SLF001
                scale_factors = to_blocked(
                    out._columnwise_scale_inv.view(torch.float8_e4m3fn),  # noqa: SLF001
                )
                amax = out._amax_columnwise  # noqa: SLF001
            else:
                e2m1_values = out._rowwise_data  # noqa: SLF001
                scale_factors = to_blocked(
                    out._rowwise_scale_inv.view(torch.float8_e4m3fn),  # noqa: SLF001
                )
                amax = out._amax_rowwise  # noqa: SLF001

        elif self == QuantizeBackend.triton:
            from .quantize.triton_kernel import quantize_to_fp4

            e2m1_values, scale_factors, amax = quantize_to_fp4(
                x,
                had=had,
                fp4_format=fp4_format,
                round_style=round_style,
                scale_rule=scale_rule,
                block_scale_2d=block_scale_2d,
                transpose=transpose,
                **kwargs,
            )

        elif self == QuantizeBackend.pytorch:
            from .quantize.reference import quantize_to_fp4

            rows_div = 128
            cols_div = 64 if fp4_format == FP4Format.nvfp4 else 128

            if x.shape[0] % rows_div != 0 or x.shape[1] % cols_div != 0:
                x = F.pad(
                    x,
                    (
                        0,
                        (
                            cols_div - (x.shape[1] % cols_div)
                            if x.shape[1] % cols_div > 0
                            else 0
                        ),
                        0,
                        (
                            rows_div - (x.shape[0] % rows_div)
                            if x.shape[0] % rows_div > 0
                            else 0
                        ),
                    ),
                )

            e2m1_values, scale_factors, amax = quantize_to_fp4(
                x,
                had=had,
                fp4_format=fp4_format,
                round_style=round_style,
                scale_rule=scale_rule,
                block_scale_2d=block_scale_2d,
                transpose=transpose,
                **kwargs,
            )

        else:
            msg = f"Invalid backend: {self}"
            raise ValueError(msg)

        return FP4Tensor(
            e2m1_values,
            scale_factors,
            amax,
            fp4_format,
            original_shape,
            scale_rule,
        )


def quantize_to_fp4(
    x: torch.Tensor,
    *,
    backend: QuantizeBackend | None = None,
    scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
    block_scale_2d: bool = False,
    had: torch.Tensor | None = None,
    fp4_format: FP4Format = FP4Format.nvfp4,
    round_style: RoundStyle = RoundStyle.nearest,
    transpose: bool = False,
) -> FP4Tensor:
    """
    Quantize a tensor to FP4.

    Args:
        x (torch.Tensor): The input tensor to quantize.
        backend (QuantizeBackend | None): The backend to use. If None, the fastest
            backend that supports the given parameters will be selected automatically.
        scale_rule (AdaptiveBlockScalingRule): The block scale selection
            rule to use, e.g. `AdaptiveBlockScalingRule.always_6` for normal NVFP4
            quantization, or `AdaptiveBlockScalingRule.mse` for 4/6 with MSE selection.
        block_scale_2d (bool): Whether to use 2D block scaling.
        had (torch.Tensor | None): The Hadamard matrix to use.
        fp4_format (FP4Format): The FP4 format to use, either `FP4Format.nvfp4` or
            `FP4Format.mxfp4`.
        round_style (RoundStyle): The rounding style to use, either `RoundStyle.nearest`
            for round-to-nearest, or `RoundStyle.stochastic` for stochastic rounding.
        transpose (bool): Whether to transpose the input tensor before quantization.

    Returns:
        A tuple containing the E2M1 values, the scale factors, and the per-tensor
            normalization constant (if used).

    """

    kwargs = {
        "scale_rule": scale_rule,
        "block_scale_2d": block_scale_2d,
        "had": had,
        "fp4_format": fp4_format,
        "round_style": round_style,
        "transpose": transpose,
    }

    if backend is None:
        backend = QuantizeBackend.auto_select(x, **kwargs)
    elif not backend.is_supported(x, **kwargs):
        msg = f"Backend {backend} does not support the given parameters"
        raise ValueError(msg)

    return backend.quantize_to_fp4(x, **kwargs)



import functools
import math

import torch

"""
Credit: TransformerEngine
https://github.com/NVIDIA/TransformerEngine/blob/main/transformer_engine/pytorch/tensor/nvfp4_tensor.py
"""


def get_no_random_sign_vector(device: int) -> torch.Tensor:
    """Non-random sign vector for Hadamard transform."""
    return torch.tensor([1], dtype=torch.float32, device=device)


def get_wgrad_sign_vector(device: int) -> torch.Tensor:
    """
    Hard-coded random signs for Hadamard transform.

    https://xkcd.com/221/

    """
    return torch.tensor(
        [1, 1, 1, -1, 1, -1, -1, -1, -1, -1, -1, 1, -1, 1, -1, -1],
        dtype=torch.float32,
        device=device,
    )


def get_hadamard_matrix(hadamard_dimension: int, device: int) -> torch.Tensor:
    """Construct a 16x16 Hadamard matrix."""
    assert hadamard_dimension == 16, "Only hadamard dimension 16 is supported."  # noqa: S101, PLR2004
    hadamard_scale = 1 / math.sqrt(hadamard_dimension)
    return (
        torch.tensor(
            [
                [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                [1, -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, -1],
                [1, 1, -1, -1, 1, 1, -1, -1, 1, 1, -1, -1, 1, 1, -1, -1],
                [1, -1, -1, 1, 1, -1, -1, 1, 1, -1, -1, 1, 1, -1, -1, 1],
                [1, 1, 1, 1, -1, -1, -1, -1, 1, 1, 1, 1, -1, -1, -1, -1],
                [1, -1, 1, -1, -1, 1, -1, 1, 1, -1, 1, -1, -1, 1, -1, 1],
                [1, 1, -1, -1, -1, -1, 1, 1, 1, 1, -1, -1, -1, -1, 1, 1],
                [1, -1, -1, 1, -1, 1, 1, -1, 1, -1, -1, 1, -1, 1, 1, -1],
                [1, 1, 1, 1, 1, 1, 1, 1, -1, -1, -1, -1, -1, -1, -1, -1],
                [1, -1, 1, -1, 1, -1, 1, -1, -1, 1, -1, 1, -1, 1, -1, 1],
                [1, 1, -1, -1, 1, 1, -1, -1, -1, -1, 1, 1, -1, -1, 1, 1],
                [1, -1, -1, 1, 1, -1, -1, 1, -1, 1, 1, -1, -1, 1, 1, -1],
                [1, 1, 1, 1, -1, -1, -1, -1, -1, -1, -1, -1, 1, 1, 1, 1],
                [1, -1, 1, -1, -1, 1, -1, 1, -1, 1, -1, 1, 1, -1, 1, -1],
                [1, 1, -1, -1, -1, -1, 1, 1, -1, -1, 1, 1, 1, 1, -1, -1],
                [1, -1, -1, 1, -1, 1, 1, -1, -1, 1, 1, -1, 1, -1, -1, 1],
            ],
            dtype=torch.float32,
            device=device,
        )
        * hadamard_scale
    )


@functools.cache
def get_rht_matrix(
    *,
    with_random_sign_mask: bool = True,
    device: str | int = "cuda",
) -> torch.Tensor:
    """Construct matrix used in random Hadamard transform."""
    hadamard_dimension = 16
    if with_random_sign_mask:
        signs = get_wgrad_sign_vector(device=device)
    else:
        signs = get_no_random_sign_vector(device=device)
    sign_matrix = signs * torch.eye(
        hadamard_dimension,
        dtype=torch.float32,
        device=device,
    )
    rht_matrix = sign_matrix @ get_hadamard_matrix(hadamard_dimension, device=device)
    return rht_matrix.to(dtype=torch.bfloat16)



import torch

import fouroversix._C  # noqa: F401


def gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,
    b_sf: torch.Tensor,
    alpha: torch.Tensor,
) -> torch.Tensor:
    return torch.ops.fouroversix.gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt.default(
        a,
        b,
        a_sf,
        b_sf,
        alpha,
    )


@torch.library.register_fake("fouroversix::gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt")
def _(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,  # noqa: ARG001
    b_sf: torch.Tensor,  # noqa: ARG001
    alpha: torch.Tensor,  # noqa: ARG001
) -> torch.Tensor:
    m = a.shape[0]
    n = b.shape[0]
    return torch.empty(m, n, dtype=torch.bfloat16, device=a.device)


def gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt_sm120(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,
    b_sf: torch.Tensor,
    alpha: torch.Tensor,
) -> torch.Tensor:
    return torch.ops.fouroversix.gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt_sm120.default(
        a,
        b,
        a_sf,
        b_sf,
        alpha,
    )


@torch.library.register_fake(
    "fouroversix::gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt_sm120",
)
def _(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,  # noqa: ARG001
    b_sf: torch.Tensor,  # noqa: ARG001
    alpha: torch.Tensor,  # noqa: ARG001
) -> torch.Tensor:
    m = a.shape[0]
    n = b.shape[0]
    return torch.empty(m, n, dtype=torch.bfloat16, device=a.device)


def gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,
    b_sf: torch.Tensor,
    alpha: torch.Tensor,
) -> torch.Tensor:
    return torch.ops.fouroversix.gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt.default(
        a,
        b,
        a_sf,
        b_sf,
        alpha,
    )


@torch.library.register_fake("fouroversix::gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt")
def _(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,  # noqa: ARG001
    b_sf: torch.Tensor,  # noqa: ARG001
    alpha: torch.Tensor,  # noqa: ARG001
) -> torch.Tensor:
    m = a.shape[0]
    n = b.shape[0]
    return torch.empty(m, n, dtype=torch.bfloat16, device=a.device)


def gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt_sm120(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,
    b_sf: torch.Tensor,
    alpha: torch.Tensor,
) -> torch.Tensor:
    return torch.ops.fouroversix.gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt_sm120.default(
        a,
        b,
        a_sf,
        b_sf,
        alpha,
    )


@torch.library.register_fake(
    "fouroversix::gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt_sm120",
)
def _(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,  # noqa: ARG001
    b_sf: torch.Tensor,  # noqa: ARG001
    alpha: torch.Tensor,  # noqa: ARG001
) -> torch.Tensor:
    m = a.shape[0]
    n = b.shape[0]
    return torch.empty(m, n, dtype=torch.bfloat16, device=a.device)


def gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,
    b_sf: torch.Tensor,
    alpha: torch.Tensor,
) -> torch.Tensor:
    return torch.ops.fouroversix.gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt.default(
        a,
        b,
        a_sf,
        b_sf,
        alpha,
    )


@torch.library.register_fake("fouroversix::gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt")
def _(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,  # noqa: ARG001
    b_sf: torch.Tensor,  # noqa: ARG001
    alpha: torch.Tensor,  # noqa: ARG001
) -> torch.Tensor:
    m = a.shape[0]
    n = b.shape[0]
    return torch.empty(m, n, dtype=torch.float16, device=a.device)


def gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt_sm120(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,
    b_sf: torch.Tensor,
    alpha: torch.Tensor,
) -> torch.Tensor:
    return torch.ops.fouroversix.gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt_sm120.default(
        a,
        b,
        a_sf,
        b_sf,
        alpha,
    )


@torch.library.register_fake(
    "fouroversix::gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt_sm120",
)
def _(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,  # noqa: ARG001
    b_sf: torch.Tensor,  # noqa: ARG001
    alpha: torch.Tensor,  # noqa: ARG001
) -> torch.Tensor:
    m = a.shape[0]
    n = b.shape[0]
    return torch.empty(m, n, dtype=torch.float16, device=a.device)


def quantize_to_fp4(
    x: torch.Tensor,
    is_nvfp4: bool,  # noqa: FBT001
    is_rtn: bool,  # noqa: FBT001
    is_rht: bool,  # noqa: FBT001
    is_transpose: bool,  # noqa: FBT001
    selection_rule: int,
) -> (torch.Tensor, torch.Tensor, torch.Tensor):
    return torch.ops.fouroversix.quantize_to_fp4.default(
        x,
        is_nvfp4,
        is_rtn,
        is_rht,
        is_transpose,
        selection_rule,
    )


@torch.library.register_fake("fouroversix::quantize_to_fp4")
def _(
    x: torch.Tensor,
    is_nvfp4: bool,  # noqa: FBT001
    is_rtn: bool,  # noqa: ARG001, FBT001
    is_rht: bool,  # noqa: ARG001, FBT001
    is_transpose: bool,  # noqa: ARG001, FBT001
    selection_rule: int,  # noqa: ARG001
) -> (torch.Tensor, torch.Tensor, torch.Tensor):
    return (
        torch.empty(x.shape[0], x.shape[1] // 2, dtype=torch.uint8, device=x.device),
        torch.empty(
            x.shape[0] * x.shape[1] // (16 if is_nvfp4 else 32),
            dtype=torch.float8_e4m3fn if is_nvfp4 else torch.float8_e8m0fnu,
            device=x.device,
        ),
        torch.empty(1, dtype=torch.float32, device=x.device),
    )



from importlib.metadata import version

from .backend import MatmulBackend, QuantizeBackend
from .frontend import fp4_matmul, quantize_to_fp4
from .model import quantize_model
from .quantize import FP4Tensor
from .utils import AdaptiveBlockScalingRule, DataType, FP4Format, RoundStyle

__version__ = version("fouroversix")

__all__ = [
    "AdaptiveBlockScalingRule",
    "DataType",
    "FP4Format",
    "FP4Tensor",
    "MatmulBackend",
    "QuantizeBackend",
    "RoundStyle",
    "fp4_matmul",
    "quantize_model",
    "quantize_to_fp4",
]



from enum import Enum

import torch


class AdaptiveBlockScalingRule(str, Enum):
    """
    Block scale selection rules for NVFP4 quantization.

    - `abs_max`: Between 4 and 6, select the block scale that minimizes the maximum
        absolute quantization error.
    - `always_4`: Select 4 for all blocks.
    - `always_6`: Select 6 for all blocks (normal NVFP4 quantization).
    - `l1_norm`: Between 4 and 6, select the block scale that minimizes the L1 norm of
        the quantization error.
    - `mse`: Between 4 and 6, select the block scale that minimizes the mean squared
        quantization error.
    """

    abs_max = "abs_max"
    always_4 = "always_4"
    always_6 = "always_6"
    l1_norm = "l1_norm"
    mse = "mse"

    def cuda_id(self) -> int:
        """ID for the rule in the CUDA implementation."""

        return {
            AdaptiveBlockScalingRule.abs_max: 4,
            AdaptiveBlockScalingRule.always_4: 1,
            AdaptiveBlockScalingRule.always_6: 0,
            AdaptiveBlockScalingRule.l1_norm: 2,
            AdaptiveBlockScalingRule.mse: 3,
        }[self]

    def max_allowed_e2m1_value(self) -> int:
        """Return the maximum allowed E2M1 value for the rule."""
        return 4 if self == AdaptiveBlockScalingRule.always_4 else 6

    def max_allowed_e4m3_value(self) -> int:
        """Return the maximum allowed E4M3 value for the rule."""
        return (
            448
            if self
            in {AdaptiveBlockScalingRule.always_6, AdaptiveBlockScalingRule.always_4}
            else 256
        )


class DataType(str, Enum):
    """High-precision data types."""

    auto = "auto"
    bfloat16 = "bfloat16"
    float16 = "float16"
    float32 = "float32"

    def torch(self) -> torch.dtype:
        """Return the corresponding torch.dtype."""

        if self == DataType.auto:
            return "auto"
        if self == DataType.bfloat16:
            return torch.bfloat16
        if self == DataType.float16:
            return torch.float16
        if self == DataType.float32:
            return torch.float32
        msg = f"Invalid data type: {self}"
        raise ValueError(msg)


class FP4Format(str, Enum):
    """FP4 formats."""

    mxfp4 = "mxfp4"
    nvfp4 = "nvfp4"

    def block_size(self) -> int:
        """Return the block size for the FP4 format."""

        return {
            FP4Format.mxfp4: 32,
            FP4Format.nvfp4: 16,
        }[self]

    def scale_dtype(self) -> torch.dtype:
        """Return the scale dtype for the FP4 format."""

        return {
            FP4Format.mxfp4: torch.float8_e8m0fnu,
            FP4Format.nvfp4: torch.float8_e4m3fn,
        }[self]


class RoundStyle(str, Enum):
    """
    Rounding styles for quantization.

    - `nearest`: Round to the nearest FP4 value.
    - `stochastic`: Round to the nearest FP4 value after applying random noise to each
        value.
    """

    nearest = "nearest"
    stochastic = "stochastic"



import itertools

import pytest
import torch
from fouroversix import (
    AdaptiveBlockScalingRule,
    FP4Format,
    QuantizeBackend,
    RoundStyle,
    quantize_to_fp4,
)
from fouroversix.quantize import get_rht_matrix

MSE_L1_NORM_MISMATCH_TOLERANCE = 1e-3
NUM_RANDOM_SEEDS = 10


@pytest.mark.parametrize("input_type", ["zeros", "ones", "rand01", "randn"])
@pytest.mark.parametrize(
    "input_shape",
    [(1024, 1024), (1024, 512), (512, 1024)],
)
@pytest.mark.parametrize(
    ("backend_a", "backend_b"),
    itertools.combinations(
        [
            QuantizeBackend.cuda,
            QuantizeBackend.triton,
            QuantizeBackend.pytorch,
            QuantizeBackend.transformer_engine,
        ],
        r=2,
    ),
)
@pytest.mark.parametrize("block_scale_2d", ["block_scale_2d", "no_block_scale_2d"])
@pytest.mark.parametrize("fp4_format", [FP4Format.nvfp4])
@pytest.mark.parametrize("had", ["had", "no_had"])
@pytest.mark.parametrize(
    "scale_rule",
    [
        AdaptiveBlockScalingRule.abs_max,
        AdaptiveBlockScalingRule.l1_norm,
        AdaptiveBlockScalingRule.mse,
        AdaptiveBlockScalingRule.always_4,
        AdaptiveBlockScalingRule.always_6,
    ],
)
@pytest.mark.parametrize("round_style", [RoundStyle.nearest, RoundStyle.stochastic])
@pytest.mark.parametrize("transpose", ["transpose", "no_transpose"])
def test_backend_outputs_are_consistent(
    input_type: str,
    input_shape: tuple[int, int],
    backend_a: QuantizeBackend,
    backend_b: QuantizeBackend,
    *,
    block_scale_2d: str,
    fp4_format: FP4Format,
    had: str,
    round_style: RoundStyle,
    scale_rule: AdaptiveBlockScalingRule,
    transpose: str,
) -> None:
    torch.set_printoptions(precision=10)

    if not backend_a.is_available() or not backend_b.is_available():
        pytest.skip("Backend is not available")

    block_scale_2d = block_scale_2d == "block_scale_2d"
    had = had == "had"
    transpose = transpose == "transpose"

    kwargs = {
        "block_scale_2d": block_scale_2d,
        "fp4_format": fp4_format,
        "had": get_rht_matrix() if had else None,
        "round_style": round_style,
        "scale_rule": scale_rule,
        "transpose": transpose,
    }

    if round_style == RoundStyle.stochastic:
        pytest.xfail("This test is not currently targeting stochastic rounding")

    for random_seed in range(NUM_RANDOM_SEEDS):
        print(f"Testing with random seed: {random_seed}")
        torch.manual_seed(random_seed)

        if input_type == "zeros":
            x = torch.zeros(*input_shape, dtype=torch.bfloat16, device="cuda")
        elif input_type == "ones":
            x = torch.ones(*input_shape, dtype=torch.bfloat16, device="cuda")
        elif input_type == "rand01":
            x = torch.randint(0, 2, input_shape, dtype=int, device="cuda").to(
                torch.bfloat16,
            )
        elif input_type == "randn":
            x = torch.randn(*input_shape, dtype=torch.bfloat16, device="cuda")
        else:
            msg = f"Invalid input type: {input_type}"
            raise ValueError(msg)

        if not backend_a.is_supported(x, **kwargs) or not backend_b.is_supported(
            x,
            **kwargs,
        ):
            pytest.skip("Backend is not supported")

        quantized_a = quantize_to_fp4(x, backend=backend_a, **kwargs)
        quantized_b = quantize_to_fp4(x, backend=backend_b, **kwargs)

        assert torch.allclose(quantized_a.amax, quantized_b.amax)

        sf_a = quantized_a.scale_factors.bfloat16()
        sf_b = quantized_b.scale_factors.bfloat16()

        if scale_rule in {
            AdaptiveBlockScalingRule.always_6,
            AdaptiveBlockScalingRule.always_4,
            AdaptiveBlockScalingRule.abs_max,
        }:
            assert torch.allclose(sf_a, sf_b)
            assert torch.allclose(quantized_a.e2m1_values, quantized_b.e2m1_values)
        else:
            # When computing 4/6 with the MSE and L1 norm scale rules, computing the
            # errors requires summing the errors in each block of 16 values. This
            # operation executes differently (elements are summed in different orders,
            # and floating-point addition is not associative) in PyTorch and Triton, and
            # can not be easily made deterministic in a way that allows for good
            # performance. As a result, we allow a small number of mismatches between
            # the scale factors and e2m1 values for these two rules. Fortunately,
            # abs_max does not involve a summation, so we can use it to test the
            # correctness of the rest of the 4/6 implementation.

            scale_factors_mismatch_prop = (sf_a != sf_b).sum() / sf_a.numel()
            assert scale_factors_mismatch_prop < MSE_L1_NORM_MISMATCH_TOLERANCE

            e2m1_values_mismatch_prop = (
                quantized_a.e2m1_values != quantized_b.e2m1_values
            ).sum() / quantized_a.e2m1_values.numel()
            assert e2m1_values_mismatch_prop < MSE_L1_NORM_MISMATCH_TOLERANCE


@pytest.mark.parametrize(
    "scale_rule",
    [
        AdaptiveBlockScalingRule.abs_max,
        AdaptiveBlockScalingRule.l1_norm,
        AdaptiveBlockScalingRule.mse,
        AdaptiveBlockScalingRule.always_4,
        AdaptiveBlockScalingRule.always_6,
    ],
)
def test_zeros(scale_rule: AdaptiveBlockScalingRule) -> None:
    device = "cuda" if torch.cuda.is_available() else "cpu"

    x = torch.zeros(1024, 1024, dtype=torch.bfloat16, device=device)
    x_e2m1, x_sf, x_normconst = quantize_to_fp4(
        x,
        backend=QuantizeBackend.pytorch,
        scale_rule=scale_rule,
    )

    x_e2m1_expected = torch.zeros(1024, 512, dtype=torch.uint8, device=device)
    x_sf_expected = torch.full(
        (1024 * 1024 // 16,),
        0,
        dtype=torch.float8_e4m3fn,
        device=device,
    )
    x_normconst_expected = torch.tensor(
        0,
        dtype=torch.bfloat16,
        device=device,
    )

    assert torch.allclose(x_normconst, x_normconst_expected)
    assert torch.allclose(x_sf.bfloat16(), x_sf_expected.bfloat16())
    assert torch.allclose(x_e2m1, x_e2m1_expected)



import pytest
import torch
from fouroversix import AdaptiveBlockScalingRule, FP4Format, FP4Tensor


@pytest.mark.parametrize("device", ["cpu"])
@pytest.mark.parametrize("fp4_format", [FP4Format.mxfp4])
@pytest.mark.parametrize("original_shape", [(2880, 2880)])
def test_dims(
    device: str,
    fp4_format: FP4Format,
    original_shape: tuple[int, int],
) -> None:
    torch.manual_seed(0)

    x_e2m1 = torch.randint(
        0,
        256,
        (original_shape[0], original_shape[1] // 2),
        dtype=torch.uint8,
        device=device,
    )
    x_sf = torch.randint(
        0,
        128,
        (original_shape[0], original_shape[1] // fp4_format.block_size()),
        dtype=torch.uint8,
        device=device,
    ).view(fp4_format.scale_dtype())
    x_amax = torch.full((1,), 1, device=device, dtype=torch.float32)

    FP4Tensor(
        x_e2m1,
        x_sf,
        x_amax,
        fp4_format,
        original_shape,
        AdaptiveBlockScalingRule.always_6,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import torch

from .backend import MatmulBackend, QuantizeBackend
from .utils import AdaptiveBlockScalingRule, FP4Format, RoundStyle

if TYPE_CHECKING:
    from .quantize import FP4Tensor


def fp4_matmul(
    input: torch.Tensor | FP4Tensor,
    other: torch.Tensor | FP4Tensor,
    *,
    backend: MatmulBackend | None = None,
    input_quantize_kwargs: dict[str, Any] | None = None,
    other_quantize_kwargs: dict[str, Any] | None = None,
    out_dtype: torch.dtype = torch.bfloat16,
) -> torch.Tensor:
    """
    Perform a matrix multiplication (`a @ b.T`) with two FP4-quantized tensors provided
    in row-major layout.

    ## Sample Code

    Each tensor may be provided in either high or low precision. If provided in high
    precision, tensors will be quantized to FP4 prior to the matrix multiplication, and
    quantization may be configured with the `input_quantize_kwargs` and
    `other_quantize_kwargs` parameters. For example, the following two code samples are
    equivalent:

    ### With High-Precision Inputs

    ```
    a = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    b = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    out = fp4_matmul(a, b)
    ```

    ### With Low-Precision Inputs

    ```
    a = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    b = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")

    a_quantized = quantize_to_fp4(a)
    b_quantized = quantize_to_fp4(b)

    out = fp4_matmul(a_quantized, b_quantized)
    ```

    ## Backends

    We provide two different implementations of FP4 matrix multiplication:

    - **CUTLASS**: Uses CUTLASS kernels to perform fast FP4 matrix multiplication.
        Requires a Blackwell GPU.
    - **PyTorch**: A slow implementation which dequantizes FP4 tensors, and then
        performs a high-precision matrix multiplication.

    Note that our CUTLASS kernels accumulate in FP32, so it should be roughly
    equivalent to simulations done with the PyTorch backend.

    ## Parameters

    Args:
        input (torch.Tensor | FP4Tensor): The first tensor to be multiplied.
        other (torch.Tensor | FP4Tensor): The second tensor to be multiplied.
        backend (MatmulBackend): The backend to use for the matrix multiplication,
            either `MatmulBackend.cutlass` or `MatmulBackend.pytorch`. If no backend is
            provided, CUTLASS will be used if the machine has a Blackwell GPU, and
            PyTorch will be used otherwise.
        input_quantize_kwargs (dict): If `a` is provided in high precision, these
            parameters will be passed to the `quantize_to_fp4` call done prior to the
            matrix multiplication.
        other_quantize_kwargs (dict): If `other` is provided in high precision, these
            parameters will be passed to the `quantize_to_fp4` call done prior to the
            matrix multiplication.
        out_dtype (DataType): The data type of the output tensor, either
            `DataType.bfloat16` or `DataType.float16`.

    Returns:
        The output tensor.

    """

    if input_quantize_kwargs is None:
        input_quantize_kwargs = {}

    if other_quantize_kwargs is None:
        other_quantize_kwargs = {}

    if isinstance(input, torch.Tensor):
        input = quantize_to_fp4(input, **(input_quantize_kwargs or {}))

    if isinstance(other, torch.Tensor):
        other = quantize_to_fp4(other, **(other_quantize_kwargs or {}))

    if backend is None:
        backend = MatmulBackend.auto_select()
    elif not backend.is_available():
        msg = f"Backend {backend} is not available"
        raise ValueError(msg)

    return backend.fp4_matmul(input, other, out_dtype=out_dtype)


def quantize_to_fp4(
    x: torch.Tensor,
    *,
    backend: QuantizeBackend | None = None,
    scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
    block_scale_2d: bool = False,
    had: torch.Tensor | None = None,
    fp4_format: FP4Format = FP4Format.nvfp4,
    round_style: RoundStyle = RoundStyle.nearest,
    transpose: bool = False,
) -> FP4Tensor:
    """
    Quantize a tensor to FP4.

    ## Sample Code

    ### With Four Over Six

    ```
    x = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    x_quantized = quantize_to_fp4(x)
    ```

    ### Without Four Over Six

    ```
    x = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    x_quantized = quantize_to_fp4(x, scale_rule=AdaptiveBlockScalingRule.always_6)
    ```

    ### With Stochastic Rounding

    ```
    x = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    x_quantized = quantize_to_fp4(x, round_style=RoundStyle.stochastic)
    ```

    ### With the Random Hadamard Transform

    ```
    from fouroversix.quantize import get_rht_matrix

    x = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    had = get_rht_matrix()
    x_quantized = quantize_to_fp4(x, had=had)
    ```

    ## Backends

    We provide three different implementations of FP4 quantization:

    - **CUDA**: A fast implementation written in CUDA which currently does not support
        the operations required for training (2D block scaling, stochastic rounding,
        random Hadamard transform). Requires a Blackwell GPU.
    - **Triton**: A slightly slower implementation written in Triton which supports all
        operations needed for training. Requires a Blackwell GPU.
    - **PyTorch**: A slow implementation written in PyTorch which supports all
        operations and can be run on any GPU.

    If `quantize_to_fp4` is called with `backend=None`, a backend will be selected
    automatically based on the following rules:

    - If there is no GPU available, or if the available GPU is not a Blackwell GPU,
        select PyTorch.
    - If any quantization options are set other than `scale_rule`, select Triton.
        - However, if the available GPU is SM120 (i.e. RTX 5090, RTX 6000) and
            `round_style` is set to `RoundStyle.stochastic`, select PyTorch as
            stochastic rounding does not have hardware support on SM120 GPUs.
    - Otherwise, select CUDA.

    ## Parameters

    Args:
        x (torch.Tensor): The input tensor to quantize.
        backend (QuantizeBackend): The backend to use for quantization, either
            `QuantizeBackend.cuda`, `QuantizeBackend.triton`, or
            `QuantizeBackend.pytorch`. If no backend is provided, one will be selected
            automatically based on the available GPU and the options provided. See above
            for more details.
        scale_rule (AdaptiveBlockScalingRule): The scaling rule to use during
            quantization. See (Adaptive Block Scaling)[/adaptive_block_scaling] for more
            details.
        block_scale_2d (bool): If True, scale factors will be computed across 16x16
            chunks of the input rather than 1x16 chunks. This is useful to apply to the
            weight matrix during training, so that W and W.T will be equivalent after
            quantization.
        had (torch.Tensor): A high-precision Hadamard matrix to apply to the input prior
            to quantization.
        fp4_format (FP4Format): The FP4 format to quantize to, either `FP4Format.mxfp4`
            or `FP4Format.nvfp4`.
        round_style (RoundStyle): The rounding style to apply during quantization,
            either `RoundStyle.nearest` for round-to-nearest quantization, or
            `RoundStyle.stochastic` for stochastic rounding.
        transpose (bool): If True, the output will be a quantized version of the
            transposed input. This may be helpful for certain operations during training
            as `fp4_matmul` requires that both tensors are provided in row-major format.

    Returns:
        A quantized FP4Tensor, which contains the packed E2M1 values, the FP8 scale
        factors, and the tensor-wide FP32 scale factor.

    """

    kwargs = {
        "scale_rule": scale_rule,
        "block_scale_2d": block_scale_2d,
        "had": had,
        "fp4_format": fp4_format,
        "round_style": round_style,
        "transpose": transpose,
    }

    if backend is None:
        backend = QuantizeBackend.auto_select(x, **kwargs)
    elif not backend.is_supported(x, **kwargs):
        msg = f"Backend {backend} does not support the given parameters"
        raise ValueError(msg)

    return backend.quantize_to_fp4(x, **kwargs)



import pytest
import torch
from fouroversix import (
    AdaptiveBlockScalingRule,
    QuantizeBackend,
    quantize_to_fp4,
)
from fouroversix.quantize import get_rht_matrix


@pytest.mark.parametrize("block_scale_2d", [True, False])
@pytest.mark.parametrize("rht", [True, False])
@pytest.mark.parametrize(
    "scale_rule",
    [
        AdaptiveBlockScalingRule.always_6,
        AdaptiveBlockScalingRule.always_4,
        AdaptiveBlockScalingRule.mse,
        AdaptiveBlockScalingRule.l1_norm,
        AdaptiveBlockScalingRule.abs_max,
    ],
)
def test_rht(
    *,
    block_scale_2d: bool,
    rht: bool,
    scale_rule: AdaptiveBlockScalingRule,
) -> None:
    torch.manual_seed(0)

    x = torch.randn(1024, 1024, dtype=torch.bfloat16, device="cuda")
    had = get_rht_matrix() if rht else None

    y_e2m1_identity, y_sf_identity, y_normconst_identity = quantize_to_fp4(
        x,
        had=had,
        backend=QuantizeBackend.triton,
        block_scale_2d=block_scale_2d,
        scale_rule=scale_rule,
    )
    y_e2m1_transpose, y_sf_transpose, y_normconst_transpose = quantize_to_fp4(
        x.T.contiguous(),
        had=had,
        transpose=True,
        backend=QuantizeBackend.triton,
        block_scale_2d=block_scale_2d,
        scale_rule=scale_rule,
    )

    assert torch.allclose(y_normconst_identity, y_normconst_transpose)
    assert torch.allclose(y_sf_identity.bfloat16(), y_sf_transpose.bfloat16())
    assert torch.allclose(y_e2m1_identity, y_e2m1_transpose)



from typing import Any

import torch
from fouroversix.backend import MatmulBackend, QuantizeBackend
from fouroversix.frontend import fp4_matmul, quantize_to_fp4
from fouroversix.quantize import FP4Tensor, get_rht_matrix
from fouroversix.utils import AdaptiveBlockScalingRule, FP4Format, RoundStyle
from torch import nn

HBS = 16


class FP4LinearFunction(torch.autograd.Function):
    """Differentiable FP4 linear layer."""

    @staticmethod
    def forward(
        ctx: torch.autograd.function.FunctionCtx,
        input: torch.Tensor,
        weight: torch.Tensor | FP4Tensor,
        bias: torch.Tensor = None,
        fp4_format: FP4Format = FP4Format.nvfp4,
        a_scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
        w_scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
        g_scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
        w_scale_2d: bool = False,  # noqa: FBT001, FBT002
        out_dtype: torch.dtype = torch.bfloat16,
        matmul_backend: MatmulBackend | None = None,
        quantize_backend: QuantizeBackend | None = None,
    ) -> tuple[torch.Tensor,]:
        """
        Perform an FP4 matrix multiplication. The input is provided in high precision
        and quantized to FP4 prior to the matrix multiplication, while the weight is
        provided in low precision.
        """

        if isinstance(weight, torch.Tensor):
            ctx.save_for_backward(input, weight, bias)
            weight = quantize_to_fp4(
                weight,
                backend=quantize_backend,
                block_scale_2d=w_scale_2d,
                fp4_format=fp4_format,
                scale_rule=w_scale_rule,
            )

        ctx.fp4_format = fp4_format
        ctx.a_scale_rule = a_scale_rule
        ctx.w_scale_rule = w_scale_rule
        ctx.g_scale_rule = g_scale_rule
        ctx.w_scale_2d = w_scale_2d
        ctx.out_dtype = out_dtype
        ctx.matmul_backend = matmul_backend
        ctx.quantize_backend = quantize_backend

        assert ctx.a_scale_rule == ctx.w_scale_rule  # noqa: S101

        if ctx.g_scale_rule is not None:
            assert ctx.a_scale_rule == ctx.g_scale_rule  # noqa: S101

        out = fp4_matmul(
            input.reshape(-1, input.shape[-1]),
            weight,
            backend=matmul_backend,
            input_quantize_kwargs={
                "backend": quantize_backend,
                "fp4_format": fp4_format,
                "scale_rule": a_scale_rule,
            },
            out_dtype=out_dtype,
        ).reshape(*input.shape[:-1], weight.original_shape[0])

        assert out.dtype == torch.bfloat16  # noqa: S101

        if bias is not None:
            out = out + bias

        return (out,)

    @staticmethod
    def backward(
        ctx: torch.autograd.function.FunctionCtx,
        grad_output: torch.Tensor,
    ) -> tuple[torch.Tensor, ...]:
        """Backward pass for the FP4 linear layer."""

        input, weight, bias = ctx.saved_tensors
        had = get_rht_matrix()

        assert grad_output.shape[0] == 1  # noqa: S101

        grad_input = fp4_matmul(
            grad_output[0],
            weight,
            backend=ctx.matmul_backend,
            input_quantize_kwargs={
                "backend": ctx.quantize_backend,
                "scale_rule": ctx.g_scale_rule,
                "fp4_format": ctx.fp4_format,
                "round_style": RoundStyle.stochastic,
            },
            other_quantize_kwargs={
                "backend": ctx.quantize_backend,
                "scale_rule": ctx.w_scale_rule,
                "fp4_format": ctx.fp4_format,
                "transpose": True,
                "block_scale_2d": ctx.w_scale_2d,
            },
            out_dtype=torch.bfloat16,
        ).unsqueeze(0)

        grad_weight = fp4_matmul(
            grad_output[0],
            input[0],
            backend=ctx.matmul_backend,
            input_quantize_kwargs={
                "backend": ctx.quantize_backend,
                "transpose": True,
                "round_style": RoundStyle.stochastic,
                "scale_rule": ctx.g_scale_rule,
                "fp4_format": ctx.fp4_format,
                "had": had,
            },
            other_quantize_kwargs={
                "backend": ctx.quantize_backend,
                "transpose": True,
                "scale_rule": ctx.a_scale_rule,
                "fp4_format": ctx.fp4_format,
                "had": had,
            },
            out_dtype=torch.bfloat16,
        ).unsqueeze(0)

        grad_bias = (
            grad_output.sum(0) if bias is not None and ctx.needs_input_grad[5] else None
        )

        return (
            grad_input,
            grad_weight,
            grad_bias,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
        )


class FP4Linear(nn.Linear):
    """Drop-in replacement for `nn.Linear` that uses FP4 quantization."""

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,  # noqa: FBT001, FBT002
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
        *,
        fp4_format: FP4Format = FP4Format.nvfp4,
        a_scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
        w_scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
        w_scale_2d: bool = False,
        matmul_backend: MatmulBackend | None = None,
        quantize_backend: QuantizeBackend | None = None,
        **kwargs: dict[str, Any],  # noqa: ARG002
    ) -> None:
        super().__init__(in_features, out_features, bias, device, dtype)

        self.fp4_format = fp4_format
        self.a_scale_rule = a_scale_rule
        self.w_scale_rule = w_scale_rule
        self.w_scale_2d = w_scale_2d
        self.out_dtype = torch.bfloat16
        self.matmul_backend = matmul_backend
        self.quantize_backend = quantize_backend

    def apply_ptq(self) -> None:
        """
        Prepare this layer for post-training quantization by quantizing the weight,
        storing the quantized weight, and deleting the original weight. This should not
        be done if the layer is used for training, as training requires storage of the
        high-precision weight.
        """

        weight = self.weight
        del self.weight

        self.weight = self.quantized_weight(weight)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Forward pass for the FP4 linear layer."""

        (out,) = FP4LinearFunction.apply(
            input,
            self.weight,
            self.bias,
            self.fp4_format,
            self.a_scale_rule,
            self.w_scale_rule,
            None,
            self.w_scale_2d,
            self.out_dtype,
            self.matmul_backend,
            self.quantize_backend,
        )

        return out

    def quantized_weight(self, weight: torch.Tensor | None = None) -> FP4Tensor:
        """Compute the quantized weights."""

        if isinstance(weight, torch.Tensor) or isinstance(self.weight, torch.Tensor):
            return quantize_to_fp4(
                weight if isinstance(weight, torch.Tensor) else self.weight,
                scale_rule=self.w_scale_rule,
                block_scale_2d=self.w_scale_2d,
                fp4_format=self.fp4_format,
                backend=self.quantize_backend,
            )

        return self.weight


class TrainableFP4Linear(FP4Linear):
    """
    Drop-in replacement for `nn.Linear` that uses FP4 quantization. This should be
    used instead of `FP4Linear` if the layer is used for training, as training requires
    storage of the high-precision weight and a hadamard matrix.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,  # noqa: FBT001, FBT002
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
        *,
        fp4_format: FP4Format = FP4Format.nvfp4,
        a_scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
        w_scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
        g_scale_rule: AdaptiveBlockScalingRule = AdaptiveBlockScalingRule.mse,
        matmul_backend: MatmulBackend | None = None,
        quantize_backend: QuantizeBackend | None = None,
    ) -> None:
        super().__init__(
            in_features,
            out_features,
            bias,
            device,
            dtype,
            fp4_format=fp4_format,
            a_scale_rule=a_scale_rule,
            w_scale_rule=w_scale_rule,
            w_scale_2d=True,
            matmul_backend=matmul_backend,
            quantize_backend=quantize_backend,
        )

        self.g_scale_rule = g_scale_rule

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Forward pass for the FP4 linear layer."""

        (out,) = FP4LinearFunction.apply(
            input,
            self.weight,
            self.bias,
            self.fp4_format,
            self.a_scale_rule,
            self.w_scale_rule,
            self.g_scale_rule,
            self.w_scale_2d,
            self.out_dtype,
            self.matmul_backend,
            self.quantize_backend,
        )

        return out
